@@ -1,3 +1,5 @@
+import numpy as np
+import json as js
 import sys
 from functools import wraps
 import json
@@ -7,6 +9,7 @@ import itertools
 import tempfile
 import random
 from time import sleep
+import pkg_resources
 
 from yattag import Doc
 from IPython.display import HTML,display
@@ -193,6 +196,7 @@ class TypeSpec(object):
         self.specificity = specificity
         self.suggest = suggest
 
+# Built-in type specs by column name
 type_specs = {
     'percentage': TypeSpec('Percentage', percentage_validation),
     'email_address': TypeSpec('EmailAddress', email_address_validation),
@@ -208,7 +212,7 @@ type_specs = {
     'marital_status': TypeSpec('MaritalStatus', marital_status_validation),
     'ip4': TypeSpec('Ip4', ip4_validation),
     # Lat and lon have specificities mainly to make tests deterministic:
-    # Giving them specificities 49 and 48 ensures they appear in that order 
+    # Giving them specificities 49 and 48 ensures they appear in that order
     # and after percentage, which has default specificity of 50.
     'latitude': TypeSpec('Latitude', latitude_validation, specificity=49),
     'longitude': TypeSpec('Longitude', longitude_validation, specificity=48),
@@ -230,6 +234,13 @@ type_specs = {
     'phone': TypeSpec('Phone', phone_validation, suggest=False),
     'url': TypeSpec('Url', url_validation, suggest=False)
 }
+# Invert the builtin type specs for quick access by type name
+type_specs_by_name = {
+    ts.name : ts for col, ts in type_specs.items()
+}
+# Like type_specs, but includes all columns from all
+# providers.  Computed lazily if required by validation code.
+all_type_specs = None
 
 # Type suggestions are generated for each column whose name doesn't
 # match one of our predefined column names.  Each suggestion points to
@@ -319,9 +330,234 @@ class Validation(object):
 
         return report
 
+# New validation object optimized for understandability, to replace
+# Validation.  Simply iterates through all columns of the input
+# dataframe, and for each does some checks.  The result of the checks
+# is a status object describing the column (see the classes ending in
+# "Status" below).
+class Validation2(object):
+    def __init__(self, inputs, providers, config):
+        self.config = config
+        self.init_type_specs()
+        # Create per-column statuses for all columns of input DF
+        self.statuses = []
+        for col in list(inputs):
+            if not col in all_type_specs:
+                self.statuses.append(NotRecognizedColumnStatus(col))
+            elif not is_string_dtype(inputs[col]):
+                self.statuses.append(NotAStringColumnStatus(col))
+            else:
+                regex_validation = all_type_specs[col].validation
+                validation_series = regex_validation.validate(inputs[col])
+                num_ok = validation_series.sum() # counts all Trues as 1, False as 0
+                num_total = len(inputs)
+                if num_ok == num_total:
+                    self.statuses.append(OKColumnStatus(col))
+                else:
+                    num_failed = num_total - num_ok
+                    # ahem
+                    failed_example_index = validation_series[validation_series==False].index[0]
+                    failed_example = inputs[col][failed_example_index]
+                    status = ValidationFailedColumnStatus(col, num_failed, num_total, failed_example)
+                    self.statuses.append(status)
+        # Create per-provider status
+        self.provider_statuses = []
+        for p in providers:
+            self.provider_statuses.append(self.create_provider_status(p, inputs))
+
+    # Build mapping of all columns supported by all providers to the
+    # type spec of each column.
+    def init_type_specs(self):
+        global all_type_specs
+        if all_type_specs == None:
+            all_type_specs = dict(type_specs)
+            providers = self.config.all_providers()
+            for p in providers:
+                version = p["version"]
+                if version:
+                    for input_field in version["input_fields"]:
+                        name = input_field["name"]
+                        type = input_field["type"]
+                        # Don't overwrite existing columns
+                        if not name in all_type_specs:
+                            if type in type_specs_by_name:
+                                all_type_specs[name] = type_specs_by_name[type]
+
+    def create_provider_status(self, provider_name, inputs):
+        return ProviderStatus(provider_name, inputs, self.config)
+
+    def _repr_html_(self):
+        doc, tag, text, line = Doc().ttl()
+        # Render per-column statuses
+        if (len(self.statuses) > 0):
+            with tag("table"):
+                with tag("tr"):
+                    with tag("th"):
+                        text("Column")
+                    with tag("th"):
+                        text("Status")
+                    with tag("th"):
+                        text("Description")
+                for status in self.statuses:
+                    status.render(doc, tag, text, line)
+        # Render provider statuses
+        if (len(self.provider_statuses) > 0):
+            line("h3", "Providers")
+            with tag("table"):
+                with tag("tr"):
+                    with tag("th"):
+                        text("Provider")
+                    with tag("th", style="text-align: left"):
+                        text("Status")
+                for provider_status in self.provider_statuses:
+                    provider_status.render(doc, tag, text, line)
+        return doc.getvalue()
+
+# Abstract superclass of per-column statuses.
+class ColumnStatus(object):
+    pass
+
+# If a column is not even a string, we tell the user they must convert
+# it to a string.
+class NotAStringColumnStatus(ColumnStatus):
+    def __init__(self, col):
+        self.col = col
+    def render(self, doc, tag, text, line):
+        with tag("tr"):
+            with tag("th"):
+                text(self.col)
+            with tag("td", style="background-color: lightpink"):
+                with tag("nobr"):
+                    text("Not a String Column")
+            with tag("td", style="text-align: left"):
+                text("You must convert this column to string type.")
+
+# If we don't understand the column name, we point the user to the
+# online documentation page with the list of all supported column
+# names.
+class NotRecognizedColumnStatus(ColumnStatus):
+    def __init__(self, col):
+        self.col = col
+    def render(self, doc, tag, text, line):
+        with tag("tr"):
+            with tag("th"):
+                text(self.col)
+            with tag("td", style="background-color: lightpink"):
+                with tag("nobr"):
+                    text("Unrecognized Column Name")
+            with tag("td", style="text-align: left"):
+                text("This column name is not supported. ")
+                with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#types", target="_blank"):
+                    text("Click here for a list of all supported column names.")
+
+# If one or more column values didn't pass the regex-based validation,
+# we give the the percentage of failed values.  We also point them to
+# the online documentation section for that column.
+class ValidationFailedColumnStatus(ColumnStatus):
+    def __init__(self, col, num_failed, num_total, failed_example):
+        self.col = col
+        self.num_failed = num_failed
+        self.num_total = num_total
+        self.failed_example = failed_example
+    def render(self, doc, tag, text, line):
+        if self.num_failed == self.num_total:
+            with tag("tr"):
+                with tag("th"):
+                    if self.col in type_specs:
+                        with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + self.col, target="_blank"):
+                            text(self.col)
+                    else:
+                        text(self.col)
+                with tag("td", style="background-color: lightpink"):
+                    with tag("nobr"):
+                        text("All Invalid")
+                with tag("td", style="text-align: left"):
+                    text("None of the " + str(self.num_total) +
+                         " column value(s) passed validation. ")
+                    text("One example of an invalid value is '" + str(self.failed_example) + "'. ")
+                    with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + self.col, target="_blank"):
+                        text("Click here for documentation for this column.")
+        else:
+            with tag("tr"):
+                with tag("th"):
+                    if self.col in type_specs:
+                        with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + self.col, target="_blank"):
+                            text(self.col)
+                    else:
+                        text(self.col)
+                with tag("td", style="background-color: peachpuff"):
+                    with tag("nobr"):
+                        text("Some Invalid")
+                with tag("td", style="text-align: left"):
+                    text(str(self.num_failed) + " value(s) of this column failed validation. ")
+                    text("One example of an invalid value is '" + str(self.failed_example) + "'. ")
+                    with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + self.col, target="_blank"):
+                        text("Click here for documentation for this column.")
+
+
+# If the column was recognized, and all values passed the validation,
+# we congratulate the user.
+class OKColumnStatus(ColumnStatus):
+    def __init__(self, col):
+        self.col = col
+    def render(self, doc, tag, text, line):
+        with tag("tr"):
+            with tag("th"):
+                if self.col in type_specs:
+                    with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + self.col, target="_blank"):
+                        text(self.col)
+                else:
+                    text(self.col)
+            with tag("td", style="background-color: lightgreen"):
+                with tag("nobr"):
+                    text("All Valid")
+            with tag("td", style="text-align: left"):
+                text("All values in this column are good to go.")
+
+class ProviderStatus(object):
+    def __init__(self, provider_name, inputs, config):
+        self.config = config;
+        self.provider_name = provider_name
+        self.provider = config.lookup_provider(provider_name)
+        self.inputs = inputs
+
+    def render(self, doc, tag, text, line):
+        if self.provider:
+            with tag("tr"):
+                with tag("th", style="vertical-align: top"):
+                    text(self.provider_name)
+                with tag("td", style="vertical-align: top; text-align: left"):
+                    self.render_provider_inputs(self.provider, self.inputs, doc, tag, text, line)
+        else:
+            with tag("tr"):
+                with tag("th"):
+                    text(self.provider_name)
+                with tag("td", style="background-color: lightpink"):
+                    text("Unrecognized provider name.")
+
+    def render_provider_inputs(self, provider, inputs, doc, tag, text, line):
+        cols = list(inputs)
+        if type(provider["version"]) is dict:
+            version = provider["version"]
+            for data in version["data"]:
+                for variant in data["input"]["required"]:
+                    with tag("table", style="display: inline-block; margin-right: 1em"):
+                        for input in variant:
+                            with tag("tr"):
+                                name = input["name"]
+                                if name in cols:
+                                    style = "background-color: lightgreen; text-align: left"
+                                else:
+                                    style = "background-color: #ddd; text-align: left"
+                                with tag("td", style=style):
+                                    with tag("a", href="https://docs.demyst.com/apis/python/analytics/api/#" + name, target="_blank"):
+                                        text(name)
+
 class Analytics(object):
 
-    def __init__(self, inputs={}, config_file=None, region=None, env=None, key=None, username=None, password=None):
+    def __init__(self, inputs=None, config_file=None, region=None, env=None, key=None, username=None, password=None):
+        if inputs is None:
+            inputs = {}
         self.C = load_config(config_file=config_file, region=region, env=env, key=key, username=username, password=password)
         self.key = self.C.get("API_KEY")
         self.prefix = True
@@ -338,28 +574,113 @@ class Analytics(object):
                                    inputs=inputs,
                                    shape="table", prefix=self.prefix)
 
+    def input_file(self, table_name, row_limit=None, filters=None):
+        return track_function(lambda: self.do_input_file(table_name, row_limit, filters),
+                              self.C, "API : Python : Rosefish Export")
+
+    def do_input_file(self, table_name, row_limit=None, filters=None):
+        params = { "table_name": table_name }
+        if(row_limit):
+            params["row_limit"] = row_limit
+        if(filters):
+            params["filters"] = filters
+        resp = self.C.auth_post(self.C.get("MANTA_URL") + "export_hosted_input", json=params).json()
+        link = resp["presigned_s3_get_url"]
+        return(pd.read_csv(link))
+
+    def input_files(self, headers=None):
+        return track_function(lambda: self.do_input_files(headers),
+                              self.C, "API : Python : Rosefish Product List")
+
+    def do_input_files(self, headers):
+        table_providers = self.C.auth_get(self.C.get("MANTA_URL") + "list_hosted_inputs").json()
+        if(headers):
+            filtered_providers = []
+            for provider in table_providers:
+                output_fields = provider["output_fields"]
+                output_field_names = list(map(lambda output_field: output_field["name"], output_fields))
+                if all(elem in output_field_names for elem in headers):
+                    filtered_providers.append(provider)
+        else:
+            filtered_providers = table_providers
+        return list(map(lambda provider: provider['name'], filtered_providers))
+
+    def input_file_headers(self, table_name):
+        return track_function(lambda: self.do_input_file_headers(table_name),
+                              self.C, "API : Python : Rosefish Product Headers")
+
+    def do_input_file_headers(self, table_name):
+        table_providers = self.C.auth_get(self.C.get("MANTA_URL") + "list_hosted_inputs").json()
+        provider_names = list(map(lambda provider: provider['name'], table_providers))
+        if table_name in provider_names:
+            headers = [provider["output_fields"] for provider in table_providers if provider['name'] == table_name][0]
+            header_df = pd.DataFrame(headers)
+            return(header_df.filter(items=['name', 'description', 'json_type_name']))
+        else:
+            print("Warning: could not find input file: {}".format(table_name), file=sys.stderr)
+
+    def list_stats_products(self):
+         return track_function(lambda: self.do_list_stats_products(),
+                              self.C, "API : Python : Hogfish Product List")
+
+    def do_list_stats_products(self):
+        return self.C.auth_get(self.C.get("MANTA_URL") + "hogfish_providers").json()
+
+
+    def product_stats(self, provider_names):
+        return track_function(lambda: self.do_product_stats(provider_names),
+                              self.C, "API : Python : Hogfish Stats",
+                              { "provider_names": provider_names })
+
+    def do_product_stats(self, provider_names):
+        res = []
+        for provider_name in provider_names:
+            res = res + self.call_manta_hogfish(provider_name)
+        df_res = pd.DataFrame(res)
+        if "id" in df_res:
+            df_res.drop(columns=['id'], inplace=True)
+        if "provider_id" in df_res:
+            df_res.rename(columns = {'provider_id':'product'}, inplace = True)
+        return(df_res)
+
+    def call_manta_hogfish(self, provider_name):
+        input = {
+            "table_provider": provider_name,
+        }
+        try:
+            res = self.C.auth_post(self.C.get("MANTA_URL") + "hogfish_stats", json=input)
+        except RuntimeError as e:
+            error_response = e.args[0]
+            if error_response == 'Error while making HTTP request: 400 {"error":"Stats not found"}':
+                print("Warning: stats not found for data provider: {}".format(provider_name), file=sys.stderr)
+            elif error_response == 'Error while making HTTP request: 400 {"error":"Table provider not found"}':
+                print("Warning: Data Provider {} not found".format(provider_name), file=sys.stderr)
+            return([])
+        return(res.json())
+
     def data_function_append(self, name, inputs_dframe):
         return track_function(lambda: self.do_data_function_append(name, inputs_dframe),
                               self.C, "API : Python : Data Function Append",
                               { "name": name })
 
     def do_data_function_append(self, name, inputs_dframe):
-        output = [None] * len(inputs_dframe)
-        for idx, row in progressbar(inputs_dframe.iterrows()):
+        # TODO: call async
+        non_nan_dframe = inputs_dframe.replace(np.nan, '', regex=True)
+        output = [None] * len(non_nan_dframe)
+        for idx, row in progressbar(non_nan_dframe.iterrows()):
             inputs = dict([(key, row[key]) for key in row.keys()])
             output[idx] = inputs
             post = {
-                "api_key": self.key,
                 "inputs": inputs,
                 "providers": [],
                 "data_function_id": name
             }
             channel_url = self.C.get("BLACKFIN_URL") + "v2/execute"
-            resp = requests.post(channel_url,
-                                 data = json.dumps(post),
-                                 headers={'Content-Type':'application/json'})
+            resp = self.C.auth_post(channel_url,
+                                    json=post,
+                                    headers={'Content-Type':'application/json'},
+                                    flags={"for_blackfin":True})
             body = resp.json()
-
             result = body['pipes']['data_function']['result']
             output[idx] = {**output[idx], **result}
 
@@ -414,15 +735,38 @@ class Analytics(object):
         sorted_columns = list(inputs_dframe.columns) + [x for x in appended_dframe.columns if x not in list(inputs_dframe.columns)]
         return appended_dframe[sorted_columns]
 
+    # Use this to fetch sample data for a list of providers.  The
+    # input dataframe is validated but ignored, it should have (at
+    # least) one row.
+    def fetch_sample_data(self, providers, inputs_dframe):
+        row = inputs_dframe.iloc[0,:] # Only need one row
+        inputs = dict([(key, row[key]) for key in row.keys()])
+        no_failures = self.connectors.fetch(providers, inputs=inputs, sample_mode=True)
+        if not no_failures:
+            for provider_name in providers:
+                provider_error = self.connectors.cache_get_error(provider_name, inputs)
+                if provider_error:
+                    print("Transaction with %s failed because of %s: %s. "%(provider_name, provider_error['type'], provider_error['message']), file=sys.stderr)
+        # Concat results
+        df = inputs_dframe.head(1)
+        for provider_name in providers:
+            provider_data = self.connectors.get(provider_name, '', inputs=inputs, shape="table", prefix=True)
+            if provider_data:
+                provider_df = pd.DataFrame.from_dict([provider_data])
+                df = pd.concat([df, provider_df], axis="columns")
+        return df
+
     # Easy-to-use frontend to enrich, enrich_wait, and enrich_download
-    def enrich_and_download(self, providers, inputs, validate=True):
-        id = self.enrich(providers, inputs, validate=validate)
+    def enrich_and_download(self, providers, inputs, validate=True, all_updates=False):
+        id = self.enrich(providers, inputs, validate=validate, all_updates=all_updates)
         return self.enrich_download(id)
 
-    def enrich(self, providers, inputs, validate=True):
-        return track_function(lambda: self.do_enrich(providers, inputs, validate=validate),
+    def enrich(self, providers, inputs, validate=True, all_updates=False):
+        return track_function(lambda: self.do_enrich(providers, inputs,
+                                                     validate=validate, all_updates=all_updates),
                               self.C, "API : Python : Enrich",
-                              { "provider_names": providers })
+                              { "provider_names": providers,
+                                "num_rows": len(inputs) })
 
     def enrich_credits(self, providers, inputs, validate=True):
         if (validate):
@@ -434,26 +778,26 @@ class Analytics(object):
         path = "table_providers/calculate_credit_cost?" + urlencode(url_params, doseq=True)
         return self.C.auth_get(self.C.get("MANTA_URL") + path).json()["total_cost"]
 
-    def do_enrich(self, providers, inputs, validate=True):
+    def do_enrich(self, providers, inputs, validate=True, all_updates=False):
         print("Starting enrichment...")
         if (validate):
             if (not self.validate_providers(providers, inputs)):
-                return None
+                raise RuntimeError("Validation failed")
         fd, temp_path = tempfile.mkstemp()
         try:
             inputs.to_csv(temp_path, index=False, header=False)
-            return self.enrich_csv(providers, temp_path, list(inputs))
+            return self.enrich_csv(providers, temp_path, list(inputs), all_updates)
         finally:
             os.close(fd)
 
-    def enrich_csv(self, providers, csv_file, columns):
+    def enrich_csv(self, providers, csv_file, columns, all_updates):
         dict = self.get_presigned_url_and_s3_object_key()
         url = dict["presigned_url"]
         s3_key = dict["s3_object_key"]
         self.upload_file_to_presigned_url(csv_file, url)
         region_id = self.find_region_id(self.C.get("REGION"))
         input_id = self.create_batch(csv_file, s3_key, columns, region_id)
-        run_batch_id = self.run_batch(providers, input_id, region_id)
+        run_batch_id = self.run_batch(providers, input_id, region_id, all_updates)
         print("Enrich Job ID: " + str(run_batch_id))
         return run_batch_id
 
@@ -489,7 +833,7 @@ class Analytics(object):
                 return region["id"]
         raise RuntimeError("Region not found: " + region_code)
 
-    def run_batch(self, providers, input_id, region_id):
+    def run_batch(self, providers, input_id, region_id, all_updates):
         provider_ids = [self.C.provider_name_to_version_id(name) for name in providers]
         # Send draft request to get org credits info
         input = {
@@ -501,6 +845,8 @@ class Analytics(object):
                 "draft": True,
             }
         }
+        if all_updates is True or all_updates is False:
+            input["aegean_batch_run"]["all_updates"] = all_updates
         draft_res = self.C.auth_post(self.C.get("MANTA_URL") + "aegean_batch_runs",
                                      json=input).json()
         org_credits = draft_res["organization"]["credit_balance"]
@@ -584,6 +930,11 @@ class Analytics(object):
         return ", ".join(messages)
 
     def enrich_download(self, id, block_until_complete=True):
+        return track_function(lambda: self.do_enrich_download(id, block_until_complete=block_until_complete),
+                              self.C, "API : Python : Download",
+                              { "job_id": id })
+
+    def do_enrich_download(self, id, block_until_complete=True):
         if block_until_complete:
             self.enrich_wait(id)
         resp_json = self.C.auth_get(self.C.get("MANTA_URL") + "aegean_batch_runs/" + str(id)).json()
@@ -801,6 +1152,13 @@ class Analytics(object):
 
             return HTML(doc.getvalue())
 
+    def validate2(self, inputs=None, providers=None):
+        if inputs is None:
+            inputs = pd.DataFrame({})
+        if providers is None:
+            providers = []
+        return Validation2(inputs, providers, self.C)
+
     def credits(self):
         headers = { "Accept": "application/json" }
         resp_json = self.C.auth_get(self.C.get("MANTA_URL") + "organization", headers=headers).json()
@@ -813,8 +1171,21 @@ class Analytics(object):
 
     # Product catalog: simple JSON info about available providers
 
-    def products(self, full=False):
-        return self.connectors.products(full=full)
+    def products(self, params=None, full=False):
+        if params is None:
+            params = {}
+        connectors = self.connectors.products(full=full)
+        if("stats_available" in params):
+            stats_products = self.list_stats_products()
+            stats_connectors = [connector for connector in connectors if connector['name'] in stats_products]
+            return(stats_connectors)
+        else:
+            return(connectors)
 
     def product_catalog(self, provider_name):
         return self.connectors.product_catalog(provider_name)
+
+def show_package_versions():
+    installed_packages = pkg_resources.working_set
+    for i in installed_packages:
+        print("%s==%s" % (i.key, i.version))
